using System;
using System.Collections.Generic;
using System.Configuration;
using System.Linq;
using SimpleJSON;
using MacGruber;
using MeshVR;
using MVR.FileManagementSecure;
using UnityEngine;

namespace CheesyFX
{
    public class FlexIt : MVRScript
    {
        private static string morphPath = "Custom/Atom/Person/Morphs/female/CheesyFX/FlexIt/";
        private const string saveDir = "Saves/PluginData/CheesyFX/FlexIt/";
        private static JSONClass factoryDefaults;
        private List<FlexRegion> flexRegions = new List<FlexRegion>();
        private static float _speed = 10f;
        public static float flexRegressionSpeed  = _speed * Time.fixedDeltaTime;
        
        private static GenerateDAZMorphsControlUI morphUI;
        private UIDynamicTabBar regionTabbar;
        public JSONStorableUrl loadURL;
        
        private JSONStorableBool regionEnabled = new JSONStorableBool("Region Enabled", true);
        private UIDynamicToggle regionEnabledUid;
        private JSONStorableStringChooser renderOption = new JSONStorableStringChooser("renderOption",
            new List<string> {"Realtime", "30", "36", "40", "45", "60", "72", "80", "90", "120", "144", "165", "240", "288"}, "Realtime", "Renderer Fps");
        
        private string[] regionNames = { "Ankle", "Knee", "Thigh", "Abdomen", "Shoulder", "Elbow"};
        private string[][] rbNames = {new[]{"Foot"}, new[]{"Shin"}, new[]{ "Thigh" }, new[]{ "pelvis", "chest" }, new[]{ "Shldr" }, new[]{ "ForeArm" }};
        
        public static bool isRealtime = true;
        public static float _targetFrametime = 1f / 60f;
        private static float _realFrametime;
        
        public static float realFrametime => isRealtime ? Time.fixedDeltaTime : _realFrametime;
        public static float targetFrametime => isRealtime ? Time.fixedDeltaTime : _targetFrametime;
        public static float physicsRateFactor = 1f;

        private static System.Diagnostics.Stopwatch watch = new System.Diagnostics.Stopwatch();

        public void OnEnable()
        {
            for (int i = 0; i < flexRegions.Count; i++)
            {
                flexRegions[i].enabled = flexRegions[i].enabledJ.val;
            }
            regionEnabled.toggle.interactable = true;
            if(regionEnabledUid) regionEnabledUid.label = "Region Enabled";
            ShakeIt.SetupTension(true);
        }

        public override void Init()
        {
            FileManagerSecure.CreateDirectory(saveDir);
            DAZCharacterSelector dcs = containingAtom.GetComponentInChildren<DAZCharacterSelector>();
            if (dcs.selectedCharacter.isMale)
            {
                morphPath = morphPath.Replace("female", "male");
            }
            // factoryDefaults = "Custom/Scripts/CheesyFX/FlexIt/FactoryDefaults.json";
            if (Utils.IsInPackage(this))
            {
                var packageUid = Utils.GetPluginPath(this).Split(':')[0] + ":/";
                // factoryDefaults = packageUid + factoryDefaults;
                morphPath = packageUid + morphPath;
            }
            morphUI = dcs.morphsControlUI;

            // rtoeColliding.setCallbackFunction += val => val.Print();
            // rToeColliding.setCallbackFunction += val => val.Print();
            
            
            for(int i=0; i<regionNames.Length; i++)
            {
                var flexRegion = gameObject.AddComponent<FlexRegion>();
                flexRegion.Init(i, regionNames[i], rbNames[i], i!=3);
                flexRegions.Add(flexRegion);
            }
            SetupHierarchy();
            
            loadURL = new JSONStorableUrl("loadURL", "", UILoadJSON, "json", true);
            loadURL.hideExtension = true;
            loadURL.setCallbackFunction -= UILoadJSON;
            loadURL.allowFullComputerBrowse = false;
            loadURL.allowBrowseAboveSuggestedPath = true;
            loadURL.SetFilePath(saveDir);
            loadURL.setCallbackFunction += UILoadJSON;
            
            UIManager.OnInitUI(CreateUIElement, leftUIElements, rightUIElements);
            Utils.OnInitUI(CreateUIElement);
            CreateUI();
            factoryDefaults = Store();
            ApplyLatestMatchingPreset();
        }
        
        private UIDynamicV4Slider sRatioUid;
        private UIDynamicV4Slider aStrengthUid;
        private UIDynamicV4Slider gStrengthUid;
        private UIDynamicV4Slider minAccelUid;
        private UIDynamicV4Slider maxMorphUid;
        private UIDynamicTVector4Legend v4legend;
        private void CreateUI()
        {
            var twinButton = Utils.SetupTwinButton(this,
                "Save preset", UISaveJSONDialog,
                "Load preset", delegate {}, false);
            loadURL.RegisterFileBrowseButton(twinButton.buttonRight);
            Utils.SetupTwinButton(this,
                "UserDefaults", () => ApplyFromUrl(saveDir+"UserDefaults.json"),
                "FactoryDefaults", () => Load(factoryDefaults), true);
            
            regionTabbar = UIManager.CreateTabBar(regionNames, val => OnRegionChanged(), script:this, invokeDefault:false);

            v4legend = UIManager.CreateVector4Legend(this);
            maxMorphUid = UIManager.CreateV4Slider(script: this, fontSize:25);
            sRatioUid = UIManager.CreateV4Slider(script: this, constrained:true, fontSize:25);
            aStrengthUid = UIManager.CreateV4Slider(script: this, fontSize:25);

            System.Text.StringBuilder sb = new System.Text.StringBuilder();
            sb.AppendLine("<b>Amount:</b>");
            sb.AppendLine("Sets the maximum morph value achievable and scales the dynamic effects.");
            sb.AppendLine("");
            sb.AppendLine("<b>S-Ratio:</b>");
            sb.AppendLine("Sets which portion of the max morph value can be achieved in static poses. If <1, full morphing is only achievable through dynamic effects (angular acceleration).");
            sb.AppendLine("0 means only dynamic effects and 1 leads to pose based morphing only.");
            sb.AppendLine("");
            sb.AppendLine("<b>Dynamic</b>");
            sb.AppendLine("The strength of the dynamic effects. Gets multiplied with 'Amount'.");
            Utils.SetupInfoText(this, sb.ToString(), 580f, false);
            // gStrengthUid = UIManager.CreateV4Slider(script: this);
            // minAccelUid = UIManager.CreateV4Slider(script: this);
            regionEnabledUid = CreateToggle(regionEnabled, true);
            CreateScrollablePopup(renderOption, true);
            Utils.SetupInfoText(this, "Set this to the output frame rate if you want to record a video using a frame renderer that slows down time.\n" +
                                      "Otherwise leave it alone!\n\nIt affects ShakeIt as well.", 200f, true);

            renderOption.setCallbackFunction += OnRenderOptionChanged;
            regionEnabled.setCallbackFunction += val => flexRegions[regionTabbar.id].enabledJ.val = val;
            maxMorphUid.setToggleCallbackFunction += maxMorphUid.ToggleSync;
            maxMorphUid.toggle.label = "Sync";
            sRatioUid.setToggleCallbackFunction += sRatioUid.ToggleSync;
            sRatioUid.toggle.label = "Sync";
            aStrengthUid.setToggleCallbackFunction += aStrengthUid.ToggleSync;
            aStrengthUid.toggle.label = "Sync";
            
            regionTabbar.SelectTab(0);
        }

        // private void HandShakeWithHeelAdjust(MVRScript heelAdjust = null)
        // {
        //     if(heelAdjust == null)
        //     {
        //         for (int i = 0; i < transform.parent.childCount; i++)
        //         {
        //             var child = transform.parent.GetChild(i);
        //             if (child.name.EndsWith("HeelAdjust2"))
        //             {
        //                 heelAdjust = child.GetComponent<MVRScript>();
        //                 heelAdjust.SendMessage("HandShakeWithFlexIt", this, SendMessageOptions.DontRequireReceiver);
        //                 break;
        //             }
        //         }
        //         if (heelAdjust == null) return;
        //     }
        //     RegisterHeelColliders();
        // }
        
        private void HandShakeWithHeelAdjust()
        {
            flexRegions[0].UpdateSupportTrackers();
        }
        
        private void SetupHierarchy()
        {
            flexRegions[1].children = new[] { flexRegions[0] };
            flexRegions[2].children = new[] { flexRegions[1], flexRegions[0] };
        }

        private void OnRegionChanged()
        {
            var flexRegion = flexRegions[regionTabbar.id];
            switch (regionTabbar.id)
            {
                case 0:
                    v4legend.SetLabels(new []{"Straighten", "Bend", "", ""});
                    break;
                case 1:
                    v4legend.SetLabels(new []{"Straighten", "Bend", "", ""});
                    break;
                case 2:
                    v4legend.SetLabels(new []{"Forward", "Backwards", "Inwards", "Outwards"});
                    break;
                case 3:
                    v4legend.SetLabels(new []{"Crunch", "Straighten", "Crunch right", "Crunch left"});
                    break;
                case 4:
                    v4legend.SetLabels(new []{"Forward", "Backwards", "Down", "Up"});
                    break;
                case 5:
                    v4legend.SetLabels(new []{"Straighten", "Bend", "", ""});
                    break;
            }
            
            sRatioUid.RegisterVector(flexRegion.sRatioJ);
            aStrengthUid.RegisterVector(flexRegion.dynamicStrengthJ);
            maxMorphUid.RegisterVector(flexRegion.amountJ);
            
            regionEnabled.valNoCallback = flexRegion.enabledJ.val;
            
            for (int i = 0; i < 4; i++)
            {
                bool active = i < flexRegion.morphsCount;
                sRatioUid.SliderSetActive(i, active);
                aStrengthUid.SliderSetActive(i, active);
                maxMorphUid.SliderSetActive(i, active);
            }
            // flexRegion.sRatioJ.constrained = true;
            // sRatioUid.SetConstrained(true);
            // var pm = containingAtom.GetComponentInChildren<PresetManager>();
            // pm.On
        }
        
        public void FixedUpdate()
        {
            if (isRealtime)
            {
                flexRegressionSpeed = _speed * Time.fixedDeltaTime;
                physicsRateFactor = 60f * Time.fixedDeltaTime;
            }
            else
            {
                flexRegressionSpeed = _speed * targetFrametime;
                _realFrametime = (float)watch.Elapsed.TotalSeconds;
                // realFrametime.Print();
                watch.Reset();
                watch.Start();
            }
        }
        
        private void OnRenderOptionChanged(string val)
        {
            if (val == "Realtime")
            {
                isRealtime = true;
            }
            else
            {
                isRealtime = false;
                _targetFrametime = 1f / float.Parse(val);
            }
        }

        private void Reset()
        {
            for (int i = 0; i < flexRegions.Count; i++)
            {
                flexRegions[i].Reset();
            }
        }
        
        public void OnDisable()
        {
            for (int i = 0; i < flexRegions.Count; i++)
            {
                flexRegions[i].enabled = false;
            }
            regionEnabled.toggle.interactable = false;
            regionEnabledUid.label = "Region Enabled (module disabled)";
            ShakeIt.SetupTension(false);
        }

        public static float MapRange(float val, float zero, float fromMin, float fromMax, float toMin, float toMax)
        {
            float v;
            if (val >= zero) v = val < fromMax? toMax * (val - zero) / (fromMax - zero) : toMax;
            else v = val > fromMin? toMin * (val - zero) / (fromMin - zero) : toMin;
            return v;
        }
        
        public static DAZMorph GetMorphByUid(string name)
        {
            DAZMorph morph;
            string uid = morphPath + name +".vmi";
            morph = morphUI.GetMorphByUid(uid);
            // if(morph == null) $"Morph '{uid}' not found!".Print();
            return morph;
        }

        public JSONClass Store()
        {
            JSONClass jc = new JSONClass();
            for (int i = 0; i < flexRegions.Count; i++)
            {
                flexRegions[i].Store(jc);
            }
            return jc;
        }

        public void Load(JSONClass jc)
        {
            for (int i = 0; i < flexRegions.Count; i++)
            {
                flexRegions[i].Load(jc);
            }
            regionTabbar.SelectLast();
        }
        
        private void ApplyFromUrl(string path)
        {
            if(!FileManagerSecure.FileExists(path)) return;
            path = FileManagerSecure.NormalizePath(path);
            {
                JSONClass jc = LoadJSON(path).AsObject;
                Load(jc);
            }
        }
        
        public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
        {
            for (int i = 0; i < flexRegions.Count; i++)
            {
                flexRegions[i].Reset();
            }
            JSONClass jc = base.GetJSON(includePhysical, includeAppearance, true);
            jc["FlexIt"] = Store();
            return jc;
        }
        
        public override void LateRestoreFromJSON(
            JSONClass jc,
            bool restorePhysical,
            bool restoreAppearance,
            bool setMissingToDefault
        )
        {
            base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
            if (jc.HasKey("FlexIt")){
                Load(jc["FlexIt"].AsObject);
            }
        }
        
        public void UISaveJSONDialog(){
            SuperController sc = SuperController.singleton;
            sc.GetMediaPathDialog(UISaveJSON, "json", saveDir, false, true, false, null, true, null, false, false);
            sc.mediaFileBrowserUI.SetTextEntry(true);
            if(sc.mediaFileBrowserUI.fileEntryField != null){
                string filename = ShakeIt.GetSuggestedPresetName(containingAtom.name, saveDir);
                sc.mediaFileBrowserUI.fileEntryField.text = filename;
                sc.mediaFileBrowserUI.ActivateFileNameField();
            }
        }
        
        private void ApplyLatestMatchingPreset()
        {
            List<string> matchingPresetPaths = new List<string>();
            string atomName = containingAtom.name;
            string[] paths = FileManagerSecure.GetFiles(saveDir);
            for (int i = 0; i < paths.Length; i++)
            {
                string path = paths[i];
                if (!path.EndsWith(".json")) continue;
                string presetName = path.Replace(saveDir+"\\","");
                
                if(presetName.StartsWith(atomName)) matchingPresetPaths.Add(presetName);
                else
                {
                    if(ShakeIt.gender == DAZCharacterSelector.Gender.Male && presetName.StartsWith("UserDefaultsMale")) matchingPresetPaths.Insert(0, presetName);
                    else if (presetName.StartsWith("UserDefaults")) matchingPresetPaths.Insert(0, presetName);
                }
            }

            if (matchingPresetPaths.Count > 0)
            {
                string prestName = matchingPresetPaths.Last();
                ApplyFromUrl(saveDir+prestName);
                $"Successfully applied '{saveDir+prestName}'.".Print();
            }
            // else $"No matching preset for '{atomName}' found.".Print();
        }
        
        public void UILoadJSON(string url){
            JSONClass jc = LoadJSON(url).AsObject;
            if(jc != null) Load(jc);
        }
        
        public void UISaveJSON(string path)
        {
            if (string.IsNullOrEmpty(path)) return;
            string path1 = path.Replace('\\', '/')+".json";
            JSONClass jc = Store();
            // for (int i=0; i<flexRegions.Count; i++) flexRegions[i].Reset();
            SaveJSON(jc, path1);
            SuperController.singleton.DoSaveScreenshot(path+".jpg");
        }
    }
}